@@ -32,13 +32,14 @@ def curve(
     epg_lambda: Optional[Union[float, int]] = 0.01,
     epg_mu: Optional[Union[float, int]] = 0.1,
     epg_trimmingradius: Optional = np.inf,
-    epg_initnodes: Optional[int] = 2,
+    epg_extend_leaves: bool = False,
     epg_verbose: bool = False,
     device: Literal["cpu", "gpu"] = "cpu",
     plot: bool = False,
     basis: Optional[str] = "umap",
     seed: Optional[int] = None,
     copy: bool = False,
+    **kwargs,
 ):
     """\
     Generate a principal curve.
@@ -58,17 +59,14 @@ def curve(
         Choose the space to be learned by the principal tree.
     ndims_rep
         Number of dimensions to use for the inference.
-    init
-        Initialise the point positions.
     epg_lambda
         Parameter for ElPiGraph, coefficient of ‘stretching’ elasticity [Albergante20]_.
     epg_mu
         Parameter for ElPiGraph, coefficient of ‘bending’ elasticity [Albergante20]_.
     epg_trimmingradius
         Parameter for ElPiGraph, trimming radius for MSE-based data approximation term [Albergante20]_.
-    epg_initnodes
-        numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
-        in the initial step
+    epg_extend_leaves
+        Parameter for ElPiGraph, calls :func:`elpigraph.ExtendLeaves` after graph learning.
     epg_verbose
         show verbose output of epg algorithm
     device
@@ -81,6 +79,8 @@ def curve(
         A numpy random seed.
     copy
         Return a copy instead of writing to adata.
+    **kwargs
+        Arguments passsed to :func:`elpigraph.computeElasticPrincipalCurve`
     Returns
     -------
     adata : anndata.AnnData
@@ -124,14 +124,14 @@ def curve(
         Nodes,
         use_rep,
         ndims_rep,
-        init,
         epg_lambda,
         epg_mu,
         epg_trimmingradius,
-        epg_initnodes,
+        epg_extend_leaves,
         device,
         seed,
         epg_verbose,
+        **kwargs,
     )
 
     if plot:
@@ -158,12 +158,14 @@ def tree(
     epg_mu: Optional[Union[float, int]] = 0.1,
     epg_trimmingradius: Optional = np.inf,
     epg_initnodes: Optional[int] = 2,
+    epg_extend_leaves: bool = False,
     epg_verbose: bool = False,
     device: Literal["cpu", "gpu"] = "cpu",
     plot: bool = False,
     basis: Optional[str] = "umap",
     seed: Optional[int] = None,
     copy: bool = False,
+    **kwargs,
 ):
     """\
     Generate a principal tree.
@@ -217,6 +219,8 @@ def tree(
     epg_initnodes
         numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
         in the initial step
+    epg_extend_leaves
+        Parameter for ElPiGraph, calls :func:`elpigraph.ExtendLeaves` after graph learning.
     epg_verbose
         show verbose output of epg algorithm
     device
@@ -229,6 +233,8 @@ def tree(
         A numpy random seed.
     copy
         Return a copy instead of writing to adata.
+    **kwargs
+        Arguments passsed to :func:`elpigraph.computeElasticPrincipalTree`
     Returns
     -------
     adata : anndata.AnnData
@@ -246,15 +252,10 @@ def tree(
             coordinates of principal points in representation space
     """
 
-    logg.info(
-        "inferring a principal tree",
-        reset=True,
-        end=" " if settings.verbosity > 2 else "\n",
-    )
-
     adata = adata.copy() if copy else adata
 
     X, use_rep = get_data(adata, use_rep, ndims_rep)
+    X = X.values
 
     W = get_data(adata, weight_rep, ndims_rep)[0] if weight_rep is not None else None
 
@@ -292,6 +293,7 @@ def tree(
             "metrics": ppt["metric"],
             "use_rep": use_rep,
             "ndims_rep": ndims_rep,
+            "method": "ppt",
         }
 
         adata.uns["graph"] = graph
@@ -299,24 +301,27 @@ def tree(
         adata.obsm["X_R"] = ppt["R"]
 
     elif method == "epg":
-        graph, epg = tree_epg(
+        logg.info(
+            "inferring a principal tree",
+            reset=True,
+            end=" " if settings.verbosity > 2 else "\n",
+        )
+        graph, R, EPG = tree_epg(
             X,
             Nodes,
-            init,
+            use_rep,
+            ndims_rep,
             epg_lambda,
             epg_mu,
             epg_trimmingradius,
-            epg_initnodes,
+            epg_extend_leaves,
             device,
             seed,
             epg_verbose,
         )
-        graph["use_rep"] = use_rep
-        graph["ndims_rep"] = ndims_rep
-        adata.obsm["X_R"] = graph["R"]
-        del graph["R"]
         adata.uns["graph"] = graph
-        adata.uns["epg"] = epg
+        adata.uns["epg"] = EPG
+        adata.obsm["X_R"] = R
 
     if plot:
         plot_graph(adata, basis)
@@ -337,17 +342,16 @@ def circle(
     Nodes: int = None,
     use_rep: str = None,
     ndims_rep: Optional[int] = None,
-    init: Optional[DataFrame] = None,
     epg_lambda: Optional[Union[float, int]] = 0.01,
     epg_mu: Optional[Union[float, int]] = 0.1,
     epg_trimmingradius: Optional = np.inf,
-    epg_initnodes: Optional[int] = 3,
     epg_verbose: bool = False,
     device: Literal["cpu", "gpu"] = "cpu",
     plot: bool = False,
     basis: Optional[str] = "umap",
     seed: Optional[int] = None,
     copy: bool = False,
+    **kwargs,
 ):
     """\
     Generate a principal circle.
@@ -367,17 +371,12 @@ def circle(
         Choose the space to be learned by the principal tree.
     ndims_rep
         Number of dimensions to use for the inference.
-    init
-        Initialise the point positions.
     epg_lambda
         Parameter for ElPiGraph, coefficient of ‘stretching’ elasticity [Albergante20]_.
     epg_mu
         Parameter for ElPiGraph, coefficient of ‘bending’ elasticity [Albergante20]_.
     epg_trimmingradius
         Parameter for ElPiGraph, trimming radius for MSE-based data approximation term [Albergante20]_.
-    epg_initnodes
-        numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
-        in the initial step
     epg_verbose
         show verbose output of epg algorithm
     device
@@ -390,6 +389,8 @@ def circle(
         A numpy random seed.
     copy
         Return a copy instead of writing to adata.
+    **kwargs
+        Arguments passsed to :func:`elpigraph.computeElasticPrincipalCircle`
     Returns
     -------
     adata : anndata.AnnData
@@ -433,11 +434,9 @@ def circle(
         Nodes,
         use_rep,
         ndims_rep,
-        init,
         epg_lambda,
         epg_mu,
         epg_trimmingradius,
-        epg_initnodes,
         device,
         seed,
         epg_verbose,
@@ -452,14 +451,16 @@ def circle(
 def tree_epg(
     X,
     Nodes: int = None,
-    init: Optional[DataFrame] = None,
+    use_rep: str = None,
+    ndims_rep: Optional[int] = None,
     lam: Optional[Union[float, int]] = 0.01,
     mu: Optional[Union[float, int]] = 0.1,
     trimmingradius: Optional = np.inf,
-    initnodes: int = None,
+    extend_leaves: bool = False,
     device: str = "cpu",
     seed: Optional[int] = None,
     verbose: bool = True,
+    **kwargs,
 ):
 
     import elpigraph
@@ -487,101 +488,27 @@ def tree_epg(
                 "Some of the GPU dependencies are missing, use device='cpu' instead!"
             )
 
-        Tree = elpigraph.computeElasticPrincipalTree(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            GPU=True,
-            verbose=verbose,
-        )
+    EPG = elpigraph.computeElasticPrincipalTree(
+        X,
+        NumNodes=Nodes,
+        Do_PCA=False,
+        Lambda=lam,
+        Mu=mu,
+        TrimmingRadius=trimmingradius,
+        GPU=device == "gpu",
+        verbose=verbose,
+        **kwargs,
+    )[0]
 
-        R = pairwise_distances(
-            cp.asarray(X.values), cp.asarray(Tree[0]["NodePositions"])
-        )
+    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
 
-        R = cp.asnumpy(R)
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    else:
-        from .utils import cor_mat_cpu
-        from sklearn.metrics import pairwise_distances
-
-        Tree = elpigraph.computeElasticPrincipalTree(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            verbose=verbose,
-        )
-
-        R = pairwise_distances(X.values, Tree[0]["NodePositions"])
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    g = igraph.Graph(directed=False)
-    g.add_vertices(np.unique(Tree[0]["Edges"][0].flatten().astype(int)))
-    g.add_edges(
-        pd.DataFrame(Tree[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
+    graph, R, EPG = epg_to_graph(
+        EPG, X, Nodes, use_rep, ndims_rep, extend_leaves, device
     )
 
-    B = np.asarray(g.get_adjacency().data)
+    EPG["Edges"] = list(EPG["Edges"])[0]
 
-    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
-    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
-    B = B[sel, :][:, sel]
-    R = R[:, sel]
-    F = Tree[0]["NodePositions"].T[:, sel]
-    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-    tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-
-    def reconnect():
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        distmat = np.triu(pairwise_distances(F[:, tips].T))
-        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
-        distmat[distmat == 0] = np.inf
-        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
-        i, j = distmat.index[row], distmat.columns[col]
-        B[i, j] = 1
-        B[j, i] = 1
-        return B
-
-    if len(emptynodes) > 0:
-        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))
-
-    recon = len(np.unique(np.array(g.clusters().membership))) > 1
-    while recon:
-        B = reconnect()
-        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        recon = len(np.unique(np.array(g.clusters().membership))) > 1
-
-    forks = np.argwhere(np.array(g.degree()) > 2).flatten()
-
-    graph = {
-        "B": B,
-        "R": R,
-        "F": Tree[0]["NodePositions"].T,
-        "tips": tips,
-        "forks": forks,
-        "cells_fitted": X.index.tolist(),
-        "metrics": "euclidean",
-    }
-
-    Tree[0]["Edges"] = list(Tree[0]["Edges"])[0]
-
-    return graph, Tree[0]
+    return graph, R, EPG
 
 
 def curve_epg(
@@ -589,18 +516,19 @@ def curve_epg(
     Nodes: int = None,
     use_rep: str = None,
     ndims_rep: Optional[int] = None,
-    init: Optional[DataFrame] = None,
     lam: Optional[Union[float, int]] = 0.01,
     mu: Optional[Union[float, int]] = 0.1,
     trimmingradius: Optional = np.inf,
-    initnodes: int = None,
+    extend_leaves: bool = False,
     device: str = "cpu",
     seed: Optional[int] = None,
     verbose: bool = True,
+    **kwargs,
 ):
     import elpigraph
 
     X, use_rep = get_data(adata, use_rep, ndims_rep)
+    X = X.values
 
     if seed is not None:
         np.random.seed(seed)
@@ -608,109 +536,33 @@ def curve_epg(
     if device == "gpu":
         try:
             import cupy as cp
-            from .utils import cor_mat_gpu
+            from .utils import cor_mat_gpu, norm_R_cpu
             from cuml.metrics import pairwise_distances
         except ModuleNotFoundError:
             raise Exception(
                 "Some of the GPU dependencies are missing, use device='cpu' instead!"
             )
 
-        Curve = elpigraph.computeElasticPrincipalCurve(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            GPU=True,
-            verbose=verbose,
-        )
+    EPG = elpigraph.computeElasticPrincipalCurve(
+        X,
+        NumNodes=Nodes,
+        Do_PCA=False,
+        Lambda=lam,
+        Mu=mu,
+        TrimmingRadius=trimmingradius,
+        GPU=device == "gpu",
+        verbose=verbose,
+        **kwargs,
+    )[0]
 
-        R = pairwise_distances(
-            cp.asarray(X.values), cp.asarray(Curve[0]["NodePositions"])
-        )
-
-        R = cp.asnumpy(R)
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    else:
-        from .utils import cor_mat_cpu
-        from sklearn.metrics import pairwise_distances
-
-        Curve = elpigraph.computeElasticPrincipalCurve(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            verbose=verbose,
-        )
-
-        R = pairwise_distances(X.values, Curve[0]["NodePositions"])
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    g = igraph.Graph(directed=False)
-    g.add_vertices(np.unique(Curve[0]["Edges"][0].flatten().astype(int)))
-    g.add_edges(
-        pd.DataFrame(Curve[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
+    graph, R, EPG = epg_to_graph(
+        EPG, X, Nodes, use_rep, ndims_rep, extend_leaves, device
     )
 
-    B = np.asarray(g.get_adjacency().data)
-
-    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
-    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
-    B = B[sel, :][:, sel]
-    R = R[:, sel]
-    F = Curve[0]["NodePositions"].T[:, sel]
-    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-    tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-
-    def reconnect():
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        distmat = np.triu(pairwise_distances(F[:, tips].T))
-        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
-        distmat[distmat == 0] = np.inf
-        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
-        i, j = distmat.index[row], distmat.columns[col]
-        B[i, j] = 1
-        B[j, i] = 1
-        return B
-
-    if len(emptynodes) > 0:
-        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))
-
-    recon = len(np.unique(np.array(g.clusters().membership))) > 1
-    while recon:
-        B = reconnect()
-        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        recon = len(np.unique(np.array(g.clusters().membership))) > 1
-
-    forks = np.argwhere(np.array(g.degree()) > 2).flatten()
-
-    graph = {
-        "B": B,
-        "F": Curve[0]["NodePositions"].T,
-        "tips": tips,
-        "forks": forks,
-        "metrics": "euclidean",
-        "use_rep": use_rep,
-        "ndims_rep": ndims_rep,
-    }
-
-    Curve[0]["Edges"] = list(Curve[0]["Edges"])[0]
+    EPG["Edges"] = list(EPG["Edges"])
 
     adata.uns["graph"] = graph
-    adata.uns["epg"] = Curve[0]
+    adata.uns["epg"] = EPG
     adata.obsm["X_R"] = R
 
     logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
@@ -730,18 +582,18 @@ def circle_epg(
     Nodes: int = None,
     use_rep: str = None,
     ndims_rep: Optional[int] = None,
-    init: Optional[DataFrame] = None,
     lam: Optional[Union[float, int]] = 0.01,
     mu: Optional[Union[float, int]] = 0.1,
     trimmingradius: Optional = np.inf,
-    initnodes: int = None,
     device: str = "cpu",
     seed: Optional[int] = None,
     verbose: bool = True,
+    **kwargs,
 ):
     import elpigraph
 
     X, use_rep = get_data(adata, use_rep, ndims_rep)
+    X = X.values
 
     if seed is not None:
         np.random.seed(seed)
@@ -756,102 +608,23 @@ def circle_epg(
                 "Some of the GPU dependencies are missing, use device='cpu' instead!"
             )
 
-        Curve = elpigraph.computeElasticPrincipalCircle(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            GPU=True,
-            verbose=verbose,
-        )
+    EPG = elpigraph.computeElasticPrincipalCircle(
+        X,
+        NumNodes=Nodes,
+        Do_PCA=False,
+        Lambda=lam,
+        Mu=mu,
+        TrimmingRadius=trimmingradius,
+        GPU=device == "gpu",
+        verbose=verbose,
+    )[0]
 
-        R = pairwise_distances(
-            cp.asarray(X.values), cp.asarray(Curve[0]["NodePositions"])
-        )
+    graph, R, EPG = epg_to_graph(EPG, X, Nodes, use_rep, ndims_rep, False, device)
 
-        R = cp.asnumpy(R)
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    else:
-        from .utils import cor_mat_cpu
-        from sklearn.metrics import pairwise_distances
-
-        Curve = elpigraph.computeElasticPrincipalCircle(
-            X.values.astype(np.float64),
-            NumNodes=Nodes,
-            Do_PCA=False,
-            InitNodes=initnodes,
-            Lambda=lam,
-            Mu=mu,
-            TrimmingRadius=trimmingradius,
-            verbose=verbose,
-        )
-
-        R = pairwise_distances(X.values, Curve[0]["NodePositions"])
-        # Hard assigment
-        R = sparse.csr_matrix(
-            (np.repeat(1, R.shape[0]), (range(R.shape[0]), R.argmin(axis=1))), R.shape
-        ).A
-
-    g = igraph.Graph(directed=False)
-    g.add_vertices(np.unique(Curve[0]["Edges"][0].flatten().astype(int)))
-    g.add_edges(
-        pd.DataFrame(Curve[0]["Edges"][0]).astype(int).apply(tuple, axis=1).values
-    )
-
-    B = np.asarray(g.get_adjacency().data)
-
-    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
-    sel = ~np.isin(np.arange(R.shape[1]), emptynodes)
-    B = B[sel, :][:, sel]
-    R = R[:, sel]
-    F = Curve[0]["NodePositions"].T[:, sel]
-    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-    tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-
-    def reconnect():
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        distmat = np.triu(pairwise_distances(F[:, tips].T))
-        distmat = pd.DataFrame(distmat, columns=tips, index=tips)
-        distmat[distmat == 0] = np.inf
-        row, col = np.unravel_index(np.argmin(distmat.values), distmat.shape)
-        i, j = distmat.index[row], distmat.columns[col]
-        B[i, j] = 1
-        B[j, i] = 1
-        return B
-
-    if len(emptynodes) > 0:
-        logg.info("    removed %d non assigned nodes" % (len(emptynodes)))
-
-    recon = len(tips) > 0
-    while recon:
-        B = reconnect()
-        g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
-        tips = np.argwhere(np.array(g.degree()) == 1).flatten()
-        recon = len(tips) > 0
-
-    forks = np.argwhere(np.array(g.degree()) > 2).flatten()
-
-    graph = {
-        "B": B,
-        "F": F,
-        "tips": tips,
-        "forks": forks,
-        "metrics": "euclidean",
-        "use_rep": use_rep,
-        "ndims_rep": ndims_rep,
-    }
-
-    Curve[0]["Edges"] = list(Curve[0]["Edges"])[0]
+    EPG["Edges"] = list(EPG["Edges"])[0]
 
     adata.uns["graph"] = graph
-    adata.uns["epg"] = Curve[0]
+    adata.uns["epg"] = EPG
     adata.obsm["X_R"] = R
 
     logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
@@ -866,6 +639,222 @@ def circle_epg(
     return adata
 
 
+def epg_to_graph(EPG, X, Nodes, use_rep, ndims_rep, extend_leaves, device):
+    import elpigraph
+    from .utils import norm_R_cpu
+
+    if device == "gpu":
+        from cuml.metrics import pairwise_distances
+    else:
+        from sklearn.metrics import pairwise_distances
+
+    if extend_leaves:
+        EPG = elpigraph.ExtendLeaves(X, EPG, Mode="WeightedCentroid")
+        Nodes = EPG["NodePositions"].shape[0]
+
+    F = EPG["NodePositions"].T
+
+    # assign to edge and obtain R
+    R = pairwise_distances(X, EPG["NodePositions"])
+    idx_nodes = np.arange(Nodes)
+    elpigraph.utils.getProjection(X, EPG)
+    mask = [
+        np.isin(idx_nodes, EPG["Edges"][0][EPG["projection"]["edge_id"][i]]) * 1
+        for i in range(X.shape[0])
+    ]
+
+    R = R * np.vstack(mask)
+    Rsum = R.sum(axis=1)
+    norm_R_cpu(R, Rsum)
+    R = np.abs(R - 1)
+    R[R == 1] = 0
+
+    # obtain B
+    g = igraph.Graph(directed=False)
+    g.add_vertices(np.unique(EPG["Edges"][0].flatten().astype(int)))
+    g.add_edges(pd.DataFrame(EPG["Edges"][0]).astype(int).apply(tuple, axis=1).values)
+
+    B = np.asarray(g.get_adjacency().data)
+
+    emptynodes = np.argwhere(R.max(axis=0) == 0).ravel()
+
+    if len(emptynodes) > 0:
+        logg.info("    there are %d non assigned nodes" % (len(emptynodes)))
+
+    g = igraph.Graph.Adjacency((B > 0).tolist(), mode="undirected")
+    tips = np.argwhere(np.array(g.degree()) == 1).flatten()
+    forks = np.argwhere(np.array(g.degree()) > 2).flatten()
+
+    graph = {
+        "B": B,
+        "F": F,
+        "tips": tips,
+        "forks": forks,
+        "metrics": "euclidean",
+        "use_rep": use_rep,
+        "ndims_rep": ndims_rep,
+        "method": "epg",
+    }
+
+    return graph, R, EPG
+
+
+def explore_sigma(
+    adata,
+    Nodes,
+    use_rep=None,
+    ndims_rep=None,
+    sigmas=[1000, 100, 10, 1, 0.1, 0.01],
+    nsteps=1,
+    metric="euclidean",
+    seed=None,
+    plot=False,
+    second_round=False,
+    **kwargs,
+):
+    """\
+    Explore varisou sigma parameters for best tree fitting. Given that high sigma tend
+    to collapse the principal points into the middle of the whole data (meaning taking
+    in account all the datapoints regardless their locality), it is possible to explore
+    which sigma is best by detecting at which level the tree stops collapsing.
+
+    Parameters
+    ----------
+    adata
+        Annotated data matrix.
+    Nodes
+        Number of nodes composing the principial tree, use a range of 10 to 100 for
+        ElPiGraph approach and 100 to 2000 for PPT approach.
+    use_rep
+        Choose the space to be learned by the principal tree.
+    ndims_rep
+        Number of dimensions to use for the inference.
+    sigmas
+        Range of sigma parameters to test.
+    device
+        Run method on either `cpu` or on `gpu`.
+    nsteps
+        Number of SimplePPT iteration, usually 1 is enough.
+    metric
+        Distance metric to use.
+    seed
+        A numpy random seed.
+    plot
+        Plot the resulting tree.
+    second_round
+        Perform a second exploration, on a restricted sigma parameters based on the first estimated sigma.
+    **kwargs
+        Arguments passsed to :func:`elpigraph.computeElasticPrincipalCircle`
+    Returns
+    -------
+    sigma : float
+        suggested sigma value
+    """
+    import copy
+    from sklearn.metrics import pairwise_distances
+
+    X, use_rep = get_data(adata, use_rep, ndims_rep)
+
+    mindist = list()
+    verb = copy.deepcopy(settings.verbosity)
+    settings.verbosity = 0
+
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        for sigma in sigmas:
+            tree(
+                adata,
+                Nodes=Nodes,
+                use_rep=use_rep,
+                method="ppt",
+                ppt_nsteps=nsteps,
+                ppt_sigma=sigma,
+                ppt_metric=metric,
+                seed=seed,
+                **kwargs,
+            )
+            mindist.append(
+                pairwise_distances(adata.uns["graph"]["F"].T, X, metric=metric)
+                .min(axis=0)
+                .mean()
+            )
+
+    def point_on_line(a, b, p):
+        ap = p - a
+        ab = b - a
+        result = a + np.dot(ap, ab) / np.dot(ab, ab) * ab
+        return result
+
+    a = np.array([0, mindist[0]])
+    b = np.array([len(sigmas) - 1, mindist[len(sigmas) - 1]])
+    projected = list()
+
+    for i in range(len(sigmas)):
+        p = np.array([i, mindist[i]])
+
+        projected.append(point_on_line(a, b, p))
+
+    curve = np.vstack([np.arange(len(sigmas)), mindist]).T
+    proj = np.vstack(projected)
+    dists = np.array(
+        [np.linalg.norm(curve[i, :] - proj[i, :]) for i in range(len(sigmas))]
+    )
+    dists[(proj[:, 0] - curve[:, 0]) < 0] = 0
+    selected = np.argmax(dists)
+
+    if plot:
+        import matplotlib.pyplot as plt
+
+        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4), constrained_layout=True)
+        ax1.plot(range(len(sigmas)), mindist, color="k")
+        ax1.set_xticks(range(len(sigmas)))
+        ax1.set_xticklabels(sigmas)
+        ax1.scatter(selected, mindist[selected], color="r", zorder=100)
+        ax1.set_xlabel("sigma parameter")
+        ax1.set_ylabel(f"mean minimum {metric} distance")
+        ax1.grid(False)
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            tree(
+                adata,
+                Nodes=Nodes,
+                use_rep=use_rep,
+                method="ppt",
+                ppt_nsteps=3,
+                ppt_sigma=sigmas[selected],
+                ppt_lambda=1,
+                seed=seed,
+            )
+        plot_graph(
+            adata,
+            show=False,
+            size_nodes=0.1,
+            title=f"iteration {nsteps}",
+            ax=ax2,
+            forks=False,
+            tips=False,
+        )
+
+    settings.verbosity = verb
+    sigma = sigmas[selected]
+    if second_round:
+        sigmas = np.arange(sigma / 5, sigma, sigma / 5)[::-1]
+        sigma = explore_sigma(
+            adata,
+            Nodes,
+            use_rep,
+            ndims_rep,
+            sigmas,
+            nsteps,
+            metric,
+            seed,
+            plot,
+            second_round=False,
+            **kwargs,
+        )
+    return sigma
+
+
 def get_data(adata, use_rep, ndims_rep):
 
     if use_rep not in adata.obsm.keys() and f"X_{use_rep}" in adata.obsm.keys():
@@ -876,20 +865,24 @@ def get_data(adata, use_rep, ndims_rep):
         & (use_rep not in adata.obsm.keys())
         & (use_rep != "X")
     ):
-        use_rep = "X" if adata.n_vars < 50 or n_pcs == 0 else "X_pca"
-        n_pcs = None if use_rep == "X" else n_pcs
+        use_rep = "X" if adata.n_vars < 50 or ndims_rep is None else "X_pca"
+        ndims_rep = None if use_rep == "X" else ndims_rep
 
     if use_rep == "X":
         ndims_rep = None
         if sparse.issparse(adata.X):
-            X = DataFrame(adata.X.A, index=adata.obs_names)
+            X = DataFrame(adata.X.A, index=adata.obs_names, columns=adata.var_names)
         else:
-            X = DataFrame(adata.X, index=adata.obs_names)
+            X = DataFrame(adata.X, index=adata.obs_names, columns=adata.var_names)
     elif use_rep in adata.layers.keys():
         if sparse.issparse(adata.layers[use_rep]):
-            X = DataFrame(adata.layers[use_rep].A, index=adata.obs_names)
+            X = DataFrame(
+                adata.layers[use_rep].A, index=adata.obs_names, columns=adata.var_names
+            )
         else:
-            X = DataFrame(adata.layers[use_rep], index=adata.obs_names)
+            X = DataFrame(
+                adata.layers[use_rep], index=adata.obs_names, columns=adata.var_names
+            )
     elif use_rep in adata.obsm.keys():
         X = DataFrame(adata.obsm[use_rep], index=adata.obs_names)
 
