@@ -82,8 +82,17 @@ def pseudotime(
     P = pairwise_distances(graph["F"].T, metric=graph["metrics"])
 
     if n_map == 1:
-        df_l = [map_cells(graph, R=adata.obsm["X_R"], P=P, multi=False)]
+        if graph["method"] == "ppt":
+            df_l = [map_cells(graph, R=adata.obsm["X_R"], P=P, multi=False)]
+        else:
+            df_l = [map_cells_epg(graph, adata)]
     else:
+        if graph["method"] == "epg":
+            raise Exception(
+                "multiple mapping does not function with ElPiGraph method, "
+                + "use either method='ppt' in scf.tl.tree or convert R matrix into soft "
+                + "assignment matrix with scf.tl.convert_to_soft"
+            )
         if seed is not None:
             np.random.seed(seed)
             map_seeds = np.random.randint(999999999, size=n_map)
@@ -115,12 +124,7 @@ def pseudotime(
     else:
         adata.obs = pd.concat([adata.obs, df_summary], axis=1)
 
-    # list(map(lambda x: x.column))
-
-    # todict=list(map(lambda x: dict(zip(["cells"]+["_"+s for s in x.columns.tolist()],
-    #                                   [x.index.tolist()]+x.to_numpy().T.tolist())),df_l))
     names = np.arange(len(df_l)).astype(str).tolist()
-    # vals = todict
     dictionary = dict(zip(names, df_l))
     adata.uns["pseudotime_list"] = dictionary
 
@@ -341,6 +345,31 @@ def map_cells(graph, R, P, multi=False, map_seed=None):
     return df
 
 
+def map_cells_epg(graph, adata):
+    import elpigraph
+    from .graph_fitting import get_data
+
+    EPG = adata.uns["epg"]
+    ndims_rep = None if "ndims_rep" not in graph else graph["ndims_rep"]
+    X, use_rep = get_data(adata, graph["use_rep"], ndims_rep)
+    elpigraph.utils.getPseudotime(X.values, EPG, graph["root"])
+    edges = EPG["Edges"][0]
+    eid = EPG["projection"]["edge_id"]
+
+    df = pd.DataFrame(
+        {
+            "t": EPG["pseudotime"],
+            "seg": graph["pp_info"].loc[EPG["projection"]["node_id"], "seg"].values,
+            "edge": [
+                "|".join(edges[eid, :][i, :].astype(str)) for i in range(len(eid))
+            ],
+        },
+        index=adata.obs_names,
+    )
+
+    return df
+
+
 def rename_milestones(adata, new: Union[list, dict], copy: bool = False):
 
     adata = adata.copy() if copy else adata
@@ -361,3 +390,93 @@ def rename_milestones(adata, new: Union[list, dict], copy: bool = False):
     adata.obs.milestones = adata.obs.milestones.cat.rename_categories(new)
 
     return adata if copy else None
+
+
+def unroll_circle(adata: AnnData, copy: bool = False):
+    """\
+    Unroll circle to get full spectrum of pseudotime values along it.
+
+    Parameters
+    ----------
+    adata
+        Annotated data matrix.
+    copy
+        Return a copy instead of writing to adata.
+    Returns
+    -------
+    adata : anndata.AnnData
+        if `copy=True` it returns or else update fields to `adata`:
+
+        `.obs['t']`
+            assigned pseudotime value.
+        `.obs['seg']`
+            assigned segment of the tree.
+        `.obs['milestone']`
+            assigned region surrounding forks and tips.
+        `.uns['graph']['pp_seg']`
+            segments network information.
+        `.uns['graph']['pp_info']`
+            for each PP, its distance vs root and segment assignment.
+    """
+    from sklearn.metrics import pairwise_distances
+    from scipy.sparse.csgraph import shortest_path
+
+    adata = adata.copy() if copy else adata
+    pp_seg = adata.uns["graph"]["pp_seg"]
+    adata.obs.t[adata.obs.seg == "0"] = (
+        -adata.obs.t[adata.obs.seg == "0"] + adata.obs.t[adata.obs.seg == "0"].max()
+    )
+
+    adata.obs.t[adata.obs.seg == "0"] = (
+        adata.obs.t[adata.obs.seg == "0"] + adata.obs.t[adata.obs.seg == "0"].max()
+    )
+
+    adata.uns["graph"]["forks"] = np.array([], dtype=int)
+    del adata.uns["graph"]["root2"]
+    a, b = adata.uns["graph"]["root"], pp_seg["from"][0]
+    adata.uns["graph"]["tips"] = np.array([a, b])
+
+    d = 1e-6 + pairwise_distances(
+        adata.uns["graph"]["F"].T,
+        adata.uns["graph"]["F"].T,
+        metric=adata.uns["graph"]["metrics"],
+    )
+
+    to_g = adata.uns["graph"]["B"] * d
+
+    csr = csr_matrix(to_g)
+
+    g = igraph.Graph.Adjacency((to_g > 0).tolist(), mode="undirected")
+    g.es["weight"] = to_g[to_g.nonzero()]
+    root_dist_matrix = shortest_path(csr, directed=False, indices=a)
+    pp_info = pd.DataFrame(
+        {"PP": g.vs.indices, "time": root_dist_matrix, "seg": np.zeros(csr.shape[0])}
+    )
+    pp_info.seg = pp_info.seg.astype(int)
+    pp_seg = pd.DataFrame(
+        {
+            "n": 0,
+            "from": a,
+            "to": b,
+            "d": g.shortest_paths(a, b, weights="weight")[0][0],
+        },
+        index=[0],
+    )
+
+    adata.obs.seg = "0"
+    adata.obs.loc[adata.obs.t <= adata.obs.t.max() / 2, "milestones"] = str(a)
+    adata.obs.loc[adata.obs.t > adata.obs.t.max() / 2, "milestones"] = str(b)
+
+    adata.uns["graph"]["pp_info"] = pp_info
+    adata.uns["graph"]["pp_seg"] = pp_seg
+
+    logg.hint(
+        "updated\n"
+        "    .obs['t'] assigned pseudotime value.\n"
+        "    .obs['seg'] assigned segment of the tree.\n"
+        "    .uns['graph']['root'] selected root.\n"
+        "    .uns['graph']['pp_info'] for each PP, its distance vs root and segment assignment.\n"
+        "    .uns['graph']['pp_seg'] segments network information."
+    )
+
+    return adata if copy else None
